#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex>
#include <string>
#include "mpi.h"

#define DEBUG 0

typedef float_complex ElementType;

/*
  The Base class: FFT
  This class implements the provides the interface to the basic
  data elements required and the FFT operation itself.
  This is an abstract class and that means that the performFFT()
  method should be implemented by a child class before it can be
  used elsewhere.

*/

class FFT{
private:
	bool DecimationInTime;
	unsigned int NumberOfProcessors;
	unsigned int NumberOfElements;
	ElementType *InElements;
	ElementType *Result;
private:
	// Private: Get/Set Operations
	void setNumberOfProcessors(unsigned int np);

	void setNumberOfElements(unsigned int nElements);

protected:
	// Protected: Twiddle factor 
	inline ElementType TwiddleFactor(int n, int k);

public:
	// Constructor and virtual destructor
	FFT(unsigned int nElements, unsigned int nProcessors);
	virtual ~FFT();

	// The main operation: performFFT
	virtual bool performFFT()=0;

	// Get/Set operations
	bool getDecimationInTime();
	void setDecimationInTime(bool bIsDecimationInTime);

	unsigned int getNumberOfProcessors();

	unsigned int getNumberOfElements();

	ElementType* getInElements();
	void setInElements(ElementType *pInElements);
	ElementType getInElements(unsigned int index);
	void setInElements(unsigned int index, ElementType Data);

	ElementType* getResult();
	void setResult(ElementType *pResult);
	ElementType getResult(unsigned int index);
	void setResult(unsigned int index, ElementType Data);
	
	// Debug operations
	void printInElements(string header);
	void printResult(string header);
};

FFT::FFT(unsigned int nElements, unsigned int nProcessors){
	// Defaults
	this->InElements = NULL;
	this->Result = NULL;
	this->NumberOfElements = 0;

	setNumberOfProcessors (nProcessors);
	setNumberOfElements (nElements);
	setDecimationInTime (true);
}

FFT::~FFT(){
	setNumberOfElements (0);
}

inline ElementType FFT::TwiddleFactor(int n, int k){
	ElementType W;
	W = complex<float>(cos(2.0*(22.0/7.0)*(double)n*(double)k/((double)(this->NumberOfElements))),
				(-1.0)*sin(2.0*(22.0/7.0)*(double)n*(double)k/((double)(this->NumberOfElements))));
	return (W);
}

void FFT::setNumberOfProcessors(unsigned int np){
	this->NumberOfProcessors = np;
}

void FFT::setNumberOfElements(unsigned int nElements){
	// Free the old buffer, if we had any
	if (this->Result != this->InElements){
		// Both are pointing to different buffers
		// So, free them separately
		if (this->Result != NULL){
			delete this->Result;
			this->Result = NULL;
		}
		if (this->InElements != NULL){
			delete this->InElements;
			this->InElements = NULL;
		}
	}
	else{
		// Both are pointing to the same buffer
		// Just free only once
		if (this->Result != NULL){
			delete this->Result;
			this->Result = NULL;
			this->InElements = NULL;
		}
	}

	// Copy the new number of elements
	this->NumberOfElements = nElements;

	// Now, allocate memory as needed
	// Default: Result points to the InElements array
	if (nElements > 0){
		this->InElements = new ElementType[nElements];
		this->Result = this->InElements;
	}
}

bool FFT::getDecimationInTime(){
	return (this->DecimationInTime);
}

void FFT::setDecimationInTime(bool bIsDecimationInTime){
	this->DecimationInTime = bIsDecimationInTime;
}

unsigned int FFT::getNumberOfProcessors(){
	return (this->NumberOfProcessors);
}

unsigned int FFT::getNumberOfElements(){
	return (this->NumberOfElements);
}

ElementType* FFT::getInElements(){
	return (this->InElements);
}

void FFT::setInElements(ElementType *pInElements){
	if (pInElements!=NULL && getNumberOfElements()>0){
		if (this->InElements == this->Result){
			if (this->InElements != NULL){
				delete this->InElements;
			}
			this->InElements = this->Result = pInElements;
		}
		else{
			if (this->InElements != NULL){
				delete this->InElements;
			}
			this->InElements = pInElements;
		}
	}
}

ElementType FFT::getInElements(unsigned int index){
	ElementType Data;
	if (index < getNumberOfElements()  &&  getInElements() != NULL){
		Data = InElements[index];
	}
	return (Data);
}

void FFT::setInElements(unsigned int index, ElementType Data){
	if (index < getNumberOfElements()  &&  getInElements() != NULL){
		InElements[index] = Data;
	}
}

ElementType* FFT::getResult(){
	return (Result);
}

void FFT::setResult(ElementType *pResult){
	if (pResult!=NULL && getNumberOfElements()>0){
		this->InElements = pResult;
	}
}
	
ElementType FFT::getResult(unsigned int index){
	ElementType Data;
	if (index < getNumberOfElements()  &&  getResult() != NULL){
		Data = Result[index];
	}
	return (Data);
}

void FFT::setResult(unsigned int index, ElementType Data){
	if (index < getNumberOfElements()  &&  getResult() != NULL){
		Result[index] = Data;
	}
}

void FFT::printInElements(string header){
	cout << header << endl;
	cout << "-----------------------------------------------------" << endl;
	for (int i=0; i < (this->NumberOfElements); i++){
		cout << "[" << i << "]: " << InElements[i] << endl;
	}
	cout << endl;
}

void FFT::printResult(string header){
	cout << header << endl;
	cout << "-----------------------------------------------------" << endl;
	for (int i=0; i < (this->NumberOfElements); i++){
		cout << "[" << i << "]: " << Result[i] << endl;
	}
	cout << endl;
}


/*
	The Derived Class: Radix2FFT derived from FFT
	This class implements the Radix-2 FFT algorithm
	This has both the uniprocessor version and the MPI version for
	multi-processors.
*/
class Radix2FFT: public FFT{
private:
	void BitReverse(ElementType *Data);
	void TwoPointFFT(unsigned int index1, unsigned int index2, unsigned int k);
	bool uniProcessorFFT();
	unsigned int LayOutCommMap(unsigned int np, bool BinTree, int **SendMap, int **RecvMap, int **StartFrom, int **EndBefore);
	bool mpiFFT();
	bool mpiFFT2();
public:
	// Constructor and virtual destructor
	Radix2FFT(unsigned int nElements, unsigned int nProcessors);
	virtual ~Radix2FFT();

	// The main operation: Radix-2 FFT : performFFT()
	bool performFFT();
};

Radix2FFT::Radix2FFT(unsigned int nElements, unsigned int nProcessors):
FFT(nElements, nProcessors)
{
}

Radix2FFT::~Radix2FFT(){
}

void Radix2FFT::BitReverse(ElementType *Data){
	// To Do: Add bit-reversal code
}

void Radix2FFT::TwoPointFFT(unsigned index1, unsigned int index2, unsigned int k){
	ElementType Input1;
	ElementType Input2;

	Input1 = getInElements(index1);
	Input2 = getInElements(index2);
	if (getDecimationInTime()==false){
		setResult (index1, Input1 + Input2);
		setResult (index2, (Input1 - Input2)*TwiddleFactor(1, k));
	}
	else{
		Input2 = Input2 * TwiddleFactor(1, k);
		setResult (index1, Input1 + Input2);
		setResult (index2, Input1 - Input2);
	}
}

bool Radix2FFT::uniProcessorFFT(){
	unsigned int nStages = 0;
	unsigned int nDivisions = 0;
	unsigned int nButterfliesPerStage = 0;
	unsigned int N = 0;
	unsigned int nElementsPerDivision = 0;
	unsigned int nButterfliesPerDivision = 0;
	unsigned int nStartFrom = 0, nEndBefore = 0, nStopBefore = 0;

	// Setup some initial values
	N = getNumberOfElements();
	nStages = (unsigned int)(ceil(log(N)/log(2)));
	nButterfliesPerStage = N/2;

	// Radix-2 FFT is performed in stages
	// In each stage we divide the entire set of input to 
	// some divisions based on the stage.
	// Then, 2-point FFTs are performed in each divisions.
	if (getDecimationInTime()==false){
		nDivisions = 1;
	}
	else{
		// Bit Reverse the input data
		BitReverse(getInElements());
		nDivisions = N / 2;
	}
	for (unsigned int i=0; i<nStages; i++){
		nElementsPerDivision = N / nDivisions;
		nButterfliesPerDivision = nElementsPerDivision / 2;
		for (unsigned int j=0; j< nDivisions; j++){
			nStartFrom = j*nElementsPerDivision;
			nEndBefore = nStartFrom + nElementsPerDivision;
			nStopBefore = nStartFrom + nButterfliesPerDivision;
			for (unsigned int k=nStartFrom; k<nStopBefore; k++){
#if(DEBUG==1)
				cout << i << ", " << j << ", " << k << ", " << k + nButterfliesPerDivision << endl;
#endif
				TwoPointFFT(k , k+nButterfliesPerDivision, (k-nStartFrom)*nDivisions);
			}
		}
		if (getDecimationInTime()==false){
			nDivisions *= 2;
		}
		else{
			nDivisions /= 2;
		}
	}
	if (getDecimationInTime()==false){
		//Reshuffle(getResult());
	}

	return true;
}

unsigned int Radix2FFT::LayOutCommMap(unsigned int np, bool BinTree, int **SendMap, int **RecvMap, int **StartFrom, int **EndBefore){
	unsigned int nStages = 0;
	unsigned int *Limit;
	unsigned int i, j;
	unsigned int k = 0;
	unsigned int N = 0;
	
	if (SendMap==NULL || RecvMap==NULL){
		return 0;
	}

	N = getNumberOfElements();
	nStages = (unsigned int)ceil(log(N)/log(2));
	Limit = new unsigned int[np];
	for (i=0; i<np; i++){
		Limit[i] = np;
	}
	for (i=0; i<nStages; i++){
		for (j=0; j<np; j++){
			StartFrom[i][j] = -1;
			EndBefore[i][j] = -1;
		}
	}
	for (i=0; i<nStages-1; i++){
		if (BinTree==false){
			// Reverse Binary Tree (with the root node in the top)
			for (j=0; j<np; j++){
				SendMap[i][j] = RecvMap [i][j] = -1;
			}
			if (i!=0){
				for (j=0; j<np; j++){
					// Default: No communication with anybody. So, just copy previous.
					StartFrom[i+1][j] = StartFrom[i][j];
					EndBefore[i+1][j] = EndBefore[i][j];
				}
				for (j=0; j<np; j++){
					if ((RecvMap[i-1][j]!=-1 || SendMap[i-1][j]!=-1) && (RecvMap[i][j]==-1 && SendMap[i][j]==-1)){
						k = j + (unsigned int)ceil((Limit[j]-j)/2);
						if (k>=Limit[j]){
							k = Limit[j] - 1;
						}
						if (k!=j){
							SendMap[i][j] = k;
							RecvMap[i][k] = j;
							Limit[k] = Limit[j];
							Limit[j] = k;
							// Calculate StartFrom, EndBefore for nStages-1-i th stage
							StartFrom[i+1][j] = StartFrom[i][j];
							EndBefore[i+1][k] = EndBefore[i][j];
							EndBefore[i+1][j] = (EndBefore[i][j]+StartFrom[i][j])/2;
							StartFrom[i+1][k] = EndBefore[i+1][j];
						}
					}
				}
			}
			else{
				k = (unsigned int)ceil(Limit[0]/2);
				if (k>=Limit[0]){
					k = Limit[0] - 1;
				}
				if (k!=0){
					SendMap[0][0] = k;
					RecvMap[0][k] = 0;
					Limit[0] = k;
					// Calculate StartFrom, EndBefore for 2nd stage
					StartFrom[1][0] = 0;
					EndBefore[1][0] = ((int)N)/2;
					StartFrom[1][k] = ((int)N)/2;
					EndBefore[1][k] = (int)N;
				}
				// StartFrom and EndBefore -> Use all components in processor 0
				StartFrom[0][0] = 0;
				EndBefore[0][0] = (int)N;
			}
		}
		else{
			// Binary Tree (with the root node in the bottom)
			for (j=0; j<np; j++){
				SendMap[nStages-2-i][j] = RecvMap [nStages-2-i][j] = -1;
			}
			if (i!=0){
				for (j=0; j<np; j++){
					// Default: No communication with anybody. So, just copy previous.
					StartFrom[nStages-2-i][j] = StartFrom[nStages-1-i][j];
					EndBefore[nStages-2-i][j] = EndBefore[nStages-1-i][j];
				}
				for (j=0; j<np; j++){
					if ((RecvMap[nStages-1-i][j]!=-1 || SendMap[nStages-1-i][j]!=-1) && (RecvMap[nStages-2-i][j]==-1 && SendMap[nStages-2-i][j]==-1)){
						k = j + (unsigned int)ceil((Limit[j]-j)/2);
						if (k>=Limit[j]){
							k = Limit[j] - 1;
						}
						if (k!=j){
							RecvMap[nStages-2-i][j] = k;
							SendMap[nStages-2-i][k] = j;
							Limit[k] = Limit[j];
							Limit[j] = k;
							// Calculate StartFrom, EndBefore for nStages-1-i th stage
							StartFrom[nStages-2-i][j] = StartFrom[nStages-1-i][j];
							EndBefore[nStages-2-i][k] = EndBefore[nStages-1-i][j];
							EndBefore[nStages-2-i][j] = (EndBefore[nStages-1-i][j]+StartFrom[nStages-1-i][j])/2;
							StartFrom[nStages-2-i][k] = EndBefore[nStages-2-i][j];
						}
					}
				}
			}
			else{
				k = (unsigned int)ceil(Limit[0]/2);
				if (k>=Limit[0]){
					k = Limit[0] - 1;
				}
				if (k!=0){
					RecvMap[nStages-2][0] = k;
					SendMap[nStages-2][k] = 0;
					Limit[0] = k;
					// Calculate StartFrom, EndBefore for nStages-1 th stage
					StartFrom[nStages-2][0] = 0;
					EndBefore[nStages-2][0] = ((int)N)/2;						
					StartFrom[nStages-2][k] = ((int)N)/2;
					EndBefore[nStages-2][k] = (int)N;						
				}
				// StartFrom and EndBefore -> Use all components i processor 0
				StartFrom[nStages-1][0] = 0;
				EndBefore[nStages-1][0] = (int)N;
			}
		}
	}

#if(DEBUG==1)
	cout << "\t StartFrom, EndBefore, Send, Recv" << endl;
	for (i=0; i<nStages-1; i++){
		cout << "Stage: " << i << endl;
		for (j=0; j<np; j++){
			if (SendMap[i][j]!=-1 || RecvMap[i][j]!=-1 || StartFrom[i][j]!=-1 || EndBefore[i][j]!=-1){
				cout << "[" << j << "]" << StartFrom[i][j] << ", " << EndBefore[i][j] << ", " << SendMap[i][j] << ", " << RecvMap[i][j] << endl;
			}
		}
	}
	cout << "Stage: " << i << endl;
	for (j=0; j<np; j++){
		if (StartFrom[i][j]!=-1 || EndBefore[i][j]!=-1){
			cout << "[" << j << "]" << StartFrom[i][j] << ", " << EndBefore[i][j] << ", -1, -1" << endl;
		}
	}
#endif
}

bool Radix2FFT::mpiFFT(){
	unsigned int nStages = 0;
	unsigned int nDivisions = 0;
	unsigned int nButterfliesPerStage = 0;
	unsigned int N = 0;
	unsigned int nElementsPerDivision = 0;
	unsigned int nButterfliesPerDivision = 0;
	unsigned int nStartFrom = 0, nEndBefore = 0, nStopBefore = 0;
	int rank;
	int np;
	unsigned int i = 0, j = 0, k = 0;
	int **SendMap = NULL, **RecvMap = NULL;
	int **StartFrom = NULL, **EndBefore = NULL;
	ElementType *Data = NULL;

	// Setup some initial values
	N = getNumberOfElements();
	nStages = (unsigned int)(ceil(log(N)/log(2)));
	nButterfliesPerStage = N/2;
	rank = MPI::COMM_WORLD.Get_rank();
	np = getNumberOfProcessors();
	Data = getInElements();

	// Allocate memory for the local arrays
	SendMap = new int* [nStages-1];
	RecvMap = new int* [nStages-1];
	StartFrom = new int* [nStages];
	EndBefore = new int* [nStages];
	StartFrom[0] = new int [np];
	EndBefore[0] = new int [np];
	for (i=0; i< nStages-1; i++){
		SendMap[i] = new int [np];
		RecvMap[i] = new int [np];
		StartFrom[1+i] = new int [np];
		EndBefore[1+i] = new int [np];
	}
	// Decide on who should communicate with whom
	LayOutCommMap(np, getDecimationInTime(), SendMap, RecvMap, StartFrom, EndBefore);

	if (getDecimationInTime()==true){
		// Wait for everybody to join us
		MPI::COMM_WORLD.Barrier();
		// Now Broadcast the input data
		MPI::COMM_WORLD.Bcast((void*)(Data), N*sizeof(ElementType)/sizeof(char), MPI::CHAR, 0);
	}

	// Radix-2 FFT is performed in stages
	// In each stage we divide the entire set of input to 
	// some divisions based on the stage.
	// Then, 2-point FFTs are performed in each divisions.
	if (getDecimationInTime()==false){
		nDivisions = 1;
	}
	else{
		// Bit Reverse the input data
		BitReverse(getInElements());
		nDivisions = N / 2;
	}
	for (i=0; i<nStages; i++){
		nElementsPerDivision = N / nDivisions;
		nButterfliesPerDivision = nElementsPerDivision / 2;
		if (StartFrom[i][rank]!=-1){
			for (j=0; j< nDivisions; j++){
				nStartFrom = j*nElementsPerDivision;
				nEndBefore = nStartFrom + nElementsPerDivision;
				nStopBefore = nStartFrom + nButterfliesPerDivision;
				if (nStartFrom>=StartFrom[i][rank] && nEndBefore<=EndBefore[i][rank]){
					for (k=nStartFrom; k<nStopBefore; k++){
#if(DEBUG==1)
						cout << "[" << rank << "]" << i << ", " << j << ", " << k << ", " << k + nButterfliesPerDivision << endl;
#endif
						TwoPointFFT(k , k+nButterfliesPerDivision, (k-nStartFrom)*nDivisions);
					}
				}
			}
		}
		if (getDecimationInTime()==false){
			nDivisions *= 2;
		}
		else{
			nDivisions /= 2;
		}
		if (i<(nStages-1)){
			if (RecvMap[i][rank]!=-1){
				MPI::COMM_WORLD.Recv((void*)(Data+StartFrom[i+1][rank]), (EndBefore[i+1][rank]-StartFrom[i+1][rank])*sizeof(ElementType)/sizeof(char), MPI::CHAR, RecvMap[i][rank], i);
			}
			else if (SendMap[i][rank]!=-1){
				int sendTo = -1;
				sendTo = SendMap[i][rank];
				MPI::COMM_WORLD.Send((const void*)(Data+StartFrom[i+1][sendTo]), (EndBefore[i+1][sendTo]-StartFrom[i+1][sendTo])*sizeof(ElementType)/sizeof(char), MPI::CHAR, sendTo, i);
			}
		}
	}
	// Wait for everybody
	MPI::COMM_WORLD.Barrier();
	// Get all the data back to proc 0 if this is decimation in frequency
	if (getDecimationInTime()==false){
		int *Count = NULL;
		int *Displ = NULL;
		Count = new int [np];
		Displ = new int [np];
		for (i=0; i<np; i++){
			if (StartFrom[nStages-1][i]==-1){
				Count[i] = 0;
				Displ[i] = 0;
			}
			else{
				Count[i] = (EndBefore[nStages-1][i]-StartFrom[nStages-1][i])*sizeof(ElementType)/sizeof(char);
				Displ[i] = StartFrom[nStages-1][i]*sizeof(ElementType)/sizeof(char);
			}
		}
		MPI::COMM_WORLD.Gatherv((const void*)(Data+StartFrom[nStages-1][rank]), Count[rank], MPI::CHAR, (void*)(Data), Count, Displ, MPI::CHAR, 0);
		//Reshuffle(getResult());
	}

	return true;
}

bool Radix2FFT::mpiFFT2(){
	unsigned int nStages = 0;
	unsigned int nDivisions = 0;
	unsigned int nButterfliesPerStage = 0;
	unsigned int N = 0;
	unsigned int nElementsPerDivision = 0;
	unsigned int nButterfliesPerDivision = 0;
	unsigned int nStartFrom = 0, nEndBefore = 0, nStopBefore = 0;
	unsigned int myStartFrom = 0, myEndBefore = 0;
	int rank;
	int np;
	unsigned int i = 0, j = 0, k = 0;
	int **SendMap = NULL, **RecvMap = NULL;
	int **StartFrom = NULL, **EndBefore = NULL;
	ElementType *Data = NULL;

	// This method is only for Decimation-in-frequency
	if (getDecimationInTime()==true){
		return false;
	}

	// Setup some initial values
	N = getNumberOfElements();
	nStages = (unsigned int)(ceil(log(N)/log(2)));
	nButterfliesPerStage = N/2;
	rank = MPI::COMM_WORLD.Get_rank();
	np = getNumberOfProcessors();
	Data = getInElements();

	// Allocate memory for the local arrays
	SendMap = new int* [nStages-1];
	RecvMap = new int* [nStages-1];
	StartFrom = new int* [nStages];
	EndBefore = new int* [nStages];
	StartFrom[0] = new int [np];
	EndBefore[0] = new int [np];
	for (i=0; i< nStages-1; i++){
		SendMap[i] = new int [np];
		RecvMap[i] = new int [np];
		StartFrom[1+i] = new int [np];
		EndBefore[1+i] = new int [np];
	}
	// Decide on who should communicate with whom
	LayOutCommMap(np, getDecimationInTime(), SendMap, RecvMap, StartFrom, EndBefore);

	// Wait for everybody to join us
	MPI::COMM_WORLD.Barrier();
	// Now Broadcast the input data
	MPI::COMM_WORLD.Bcast((void*)(Data), N*sizeof(ElementType)/sizeof(char), MPI::CHAR, 0);

	myStartFrom = StartFrom[nStages-1][rank];
	myEndBefore = EndBefore[nStages-1][rank];

	// Radix-2 FFT is performed in stages
	// In each stage we divide the entire set of input to 
	// some divisions based on the stage.
	// Then, 2-point FFTs are performed in each divisions.
	nDivisions = 1;
	for (i=0; i<nStages; i++){
		nElementsPerDivision = N / nDivisions;
		nButterfliesPerDivision = nElementsPerDivision / 2;
		if (StartFrom[i][rank]!=-1){
			for (j=0; j< nDivisions; j++){
				nStartFrom = j*nElementsPerDivision;
				nEndBefore = nStartFrom + nElementsPerDivision;
				nStopBefore = nStartFrom + nButterfliesPerDivision;
				if ((nStartFrom<=myStartFrom && myStartFrom<nEndBefore) ||
					(nStartFrom<myEndBefore && myEndBefore<=nEndBefore) ||
					(myStartFrom<=nStartFrom && nStartFrom<myEndBefore) ||
					(nStartFrom<=myStartFrom && myEndBefore<=nEndBefore)
					){
					for (k=nStartFrom; k<nStopBefore; k++){
#if(DEBUG==1)
						cout << "[" << rank << "]" << i << ", " << j << ", " << k << ", " << k + nButterfliesPerDivision << endl;
#endif
						TwoPointFFT(k , k+nButterfliesPerDivision, (k-nStartFrom)*nDivisions);
					}
				}
			}
		}
		nDivisions *= 2;
	}
	// Wait for everybody
	MPI::COMM_WORLD.Barrier();
	// Get all the data back to proc 0 if this is decimation in frequency
	int *Count = NULL;
	int *Displ = NULL;
	Count = new int [np];
	Displ = new int [np];
	for (i=0; i<np; i++){
		if (StartFrom[nStages-1][i]==-1){
			Count[i] = 0;
			Displ[i] = 0;
		}
		else{
			Count[i] = (EndBefore[nStages-1][i]-StartFrom[nStages-1][i])*sizeof(ElementType)/sizeof(char);
			Displ[i] = StartFrom[nStages-1][i]*sizeof(ElementType)/sizeof(char);
		}
	}
	MPI::COMM_WORLD.Gatherv((const void*)(Data+StartFrom[nStages-1][rank]), Count[rank], MPI::CHAR, (void*)(Data), Count, Displ, MPI::CHAR, 0);
	//Reshuffle(getResult());

	return true;
}


bool Radix2FFT::performFFT(){
	bool bDone = false;
	double timeTaken = 0;
	int rank = 0;

	rank = MPI::COMM_WORLD.Get_rank();

	if (rank==0){
		timeTaken = MPI::Wtime();
	}

	if (getNumberOfProcessors()==1){
		bDone = uniProcessorFFT();
	}
	else{
		bDone = mpiFFT2();
	}

	if (rank==0){
		timeTaken = MPI::Wtime() - timeTaken;
		cout << "Time taken for " << getNumberOfElements() << " elements using " << getNumberOfProcessors() << " processors = " << timeTaken << " seconds" << endl;
	}

	return bDone;
}

/*
	The main function
	This initialises the various part of the code and leaves the rest to the classes.
*/

int main(int argc, char **argv){
  int rank, size;
  unsigned int nElements = 16;
  FFT *fft;
  string resultStr("The spectrum is");
  string inputStr("The inputs are");

  MPI_Init(&argc, &argv);
  size = MPI::COMM_WORLD.Get_size();
  rank = MPI::COMM_WORLD.Get_rank();

  if (argc>1){
	  nElements = atoi(argv[1]);
  }
  fft = new Radix2FFT(nElements, size);
  if (rank==0){
	for (unsigned int i=0; i<nElements; i++){
		fft->setInElements(i, ElementType(rand()/nElements, rand()/nElements));
	}
  }
  //fft->printInElements(inputStr);
  fft->setDecimationInTime(false);
  fft->performFFT();
  //fft->printResult(resultStr);
  delete fft; fft = NULL;

  MPI_Finalize();
}

