#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <complex>
#include <string>
#include <list>
#include "mpi.h"
using namespace std;

#define DEBUG 0

typedef float_complex ElementType;

/*
  The Base class: FFT
  This class implements the provides the interface to the basic
  data elements required and the FFT operation itself.
  This is an abstract class and that means that the performFFT()
  method should be implemented by a child class before it can be
  used elsewhere.

*/

class FFT{
private:
	bool DecimationInTime;
	unsigned int NumberOfProcessors;
	unsigned int NumberOfElements;
	ElementType *InElements;
	ElementType *Result;
private:
	// Private: Get/Set Operations
	void setNumberOfProcessors(unsigned int np);

	void setNumberOfElements(unsigned int nElements);

protected:
	// Protected: Twiddle factor 
	inline ElementType TwiddleFactor(int n, int k);

public:
	// Constructor and virtual destructor
	FFT(unsigned int nElements, unsigned int nProcessors);
	virtual ~FFT();

	// The main operation: performFFT
	virtual bool performFFT()=0;

	// Get/Set operations
	bool getDecimationInTime();
	void setDecimationInTime(bool bIsDecimationInTime);

	unsigned int getNumberOfProcessors();

	unsigned int getNumberOfElements();

	ElementType* getInElements();
	void setInElements(ElementType *pInElements);
	ElementType getInElements(unsigned int index);
	void setInElements(unsigned int index, ElementType Data);

	ElementType* getResult();
	void setResult(ElementType *pResult);
	ElementType getResult(unsigned int index);
	void setResult(unsigned int index, ElementType Data);
	
	// Debug operations
	void printInElements(string header);
	void printResult(string header);
};

FFT::FFT(unsigned int nElements, unsigned int nProcessors){
	// Defaults
	this->InElements = NULL;
	this->Result = NULL;
	this->NumberOfElements = 0;

	setNumberOfProcessors (nProcessors);
	setNumberOfElements (nElements);
	setDecimationInTime (true);
}

FFT::~FFT(){
	setNumberOfElements (0);
}

inline ElementType FFT::TwiddleFactor(int n, int k){
	ElementType W;
	W = complex<float>(cos(2.0*(22.0/7.0)*(double)n*(double)k/((double)(this->NumberOfElements))),
				(-1.0)*sin(2.0*(22.0/7.0)*(double)n*(double)k/((double)(this->NumberOfElements))));
	return (W);
}

void FFT::setNumberOfProcessors(unsigned int np){
	this->NumberOfProcessors = np;
}

void FFT::setNumberOfElements(unsigned int nElements){
	// Free the old buffer, if we had any
	if (this->Result != this->InElements){
		// Both are pointing to different buffers
		// So, free them separately
		if (this->Result != NULL){
			delete this->Result;
			this->Result = NULL;
		}
		if (this->InElements != NULL){
			delete this->InElements;
			this->InElements = NULL;
		}
	}
	else{
		// Both are pointing to the same buffer
		// Just free only once
		if (this->Result != NULL){
			delete this->Result;
			this->Result = NULL;
			this->InElements = NULL;
		}
	}

	// Copy the new number of elements
	this->NumberOfElements = nElements;

	// Now, allocate memory as needed
	// Default: Result points to the InElements array
	if (nElements > 0){
		this->InElements = new ElementType[nElements];
		this->Result = this->InElements;
	}
}

bool FFT::getDecimationInTime(){
	return (this->DecimationInTime);
}

void FFT::setDecimationInTime(bool bIsDecimationInTime){
	this->DecimationInTime = bIsDecimationInTime;
}

unsigned int FFT::getNumberOfProcessors(){
	return (this->NumberOfProcessors);
}

unsigned int FFT::getNumberOfElements(){
	return (this->NumberOfElements);
}

ElementType* FFT::getInElements(){
	return (this->InElements);
}

void FFT::setInElements(ElementType *pInElements){
	if (pInElements!=NULL && getNumberOfElements()>0){
		if (this->InElements == this->Result){
			if (this->InElements != NULL){
				delete this->InElements;
			}
			this->InElements = this->Result = pInElements;
		}
		else{
			if (this->InElements != NULL){
				delete this->InElements;
			}
			this->InElements = pInElements;
		}
	}
}

ElementType FFT::getInElements(unsigned int index){
	ElementType Data;
	if (index < getNumberOfElements()  &&  getInElements() != NULL){
		Data = InElements[index];
	}
	return (Data);
}

void FFT::setInElements(unsigned int index, ElementType Data){
	if (index < getNumberOfElements()  &&  getInElements() != NULL){
		InElements[index] = Data;
	}
}

ElementType* FFT::getResult(){
	return (Result);
}

void FFT::setResult(ElementType *pResult){
	if (pResult!=NULL && getNumberOfElements()>0){
		this->InElements = pResult;
	}
}
	
ElementType FFT::getResult(unsigned int index){
	ElementType Data;
	if (index < getNumberOfElements()  &&  getResult() != NULL){
		Data = Result[index];
	}
	return (Data);
}

void FFT::setResult(unsigned int index, ElementType Data){
	if (index < getNumberOfElements()  &&  getResult() != NULL){
		Result[index] = Data;
	}
}

void FFT::printInElements(string header){
	cout << header << endl;
	cout << "-----------------------------------------------------" << endl;
	for (int i=0; i < (this->NumberOfElements); i++){
		cout << "[" << i << "]: " << InElements[i] << endl;
	}
	cout << endl;
}

void FFT::printResult(string header){
	cout << header << endl;
	cout << "-----------------------------------------------------" << endl;
	for (int i=0; i < (this->NumberOfElements); i++){
		cout << "[" << i << "]: " << Result[i] << endl;
	}
	cout << endl;
}

/*
	Some more support classes:
	CDataStageMap, CDataProcMap
*/
class CDataStageMap{
public:
	CDataStageMap(int startFrom, int endBefore):
	  StartFrom(startFrom),
	  EndBefore(endBefore){}
	int StartFrom;
	int EndBefore;
};
typedef list<CDataStageMap> DataStageMapList;
class CDataProcMap{
public:
	CDataProcMap(int startFrom, int endBefore, int proc):
	  StartFrom(startFrom),
	  EndBefore(endBefore),
	  Processor(proc){}
	int StartFrom;
	int EndBefore;
	int Processor;
};
typedef list<CDataProcMap> DataProcMapList;

/*
	The Derived Class: Radix4FFT derived from FFT
	This class implements the Radix-4 FFT algorithm
	This has both the uniprocessor version and the MPI version for
	multi-processors.
*/
class Radix4FFT: public FFT{
private:
	void FFTOnDivision(unsigned int startIndex, unsigned int endIndex);
	void FourPointFFT(unsigned int iIndex1, unsigned int iIndex2, unsigned int iIndex3, unsigned int iIndex4, unsigned int q);
	unsigned int LayOutDataProcMap(DataProcMapList &lastStage);
	bool mpiFFT();
	bool UniProcessorFFT();
public:
	// Constructor and virtual destructor
	Radix4FFT(unsigned int nElements, unsigned int nProcessors);
	virtual ~Radix4FFT();

	// The main operation: Radix-2 FFT : performFFT()
	bool performFFT();
};

Radix4FFT::Radix4FFT(unsigned int nElements, unsigned int nProcessors):
FFT(nElements, nProcessors)
{
}

Radix4FFT::~Radix4FFT(){
}

void Radix4FFT::FFTOnDivision(unsigned startIndex, unsigned int endIndex){
	unsigned int i = 0;
	unsigned int Bf = 0;
	ElementType  Input1, Input2;
	ElementType  *Data = NULL;

	Data = getInElements();
	Bf = (endIndex-startIndex)/4;
	for (i=startIndex; i<(startIndex+Bf); i++){
		if (getDecimationInTime()==false){
			FourPointFFT(startIndex, startIndex+Bf, startIndex+2*Bf, startIndex+3*Bf,(i-startIndex));
		}
	}
}

void Radix4FFT::FourPointFFT(unsigned int iIndex1, unsigned int iIndex2, unsigned int iIndex3, unsigned int iIndex4, unsigned int q){
	ElementType Input1, Input2, Input3, Input4;
	ElementType Temp1, Temp2, Temp3, Temp4;

	// Get Inputs
	Input1 = getResult(iIndex1);
	Input2 = getResult(iIndex2);
	Input3 = getResult(iIndex3);
	Input4 = getResult(iIndex4);

	// Outputs
	// First Stage
	Temp1 = Input1 + Input3;
	Temp2 = Input1 - Input3;
	Temp3 = Input2 + Input4;
	Temp4 = Input2 - Input4;
	Temp4 = ElementType(-Temp4.imag(), Temp4.real());
	// Second Stage
	setResult(iIndex1, (Temp1 + Temp3));
	setResult(iIndex2, (Temp2 - Temp4)*TwiddleFactor(q, 1));
	setResult(iIndex3, (Temp1 - Temp3)*TwiddleFactor(q, 2));
	setResult(iIndex4, (Temp2 + Temp4)*TwiddleFactor(q, 3));
}

unsigned int Radix4FFT::LayOutDataProcMap(DataProcMapList &lastStage){
	unsigned int nStages = 0;
	unsigned int *Limit;
	unsigned int i, j;
	unsigned int k = 0;
	unsigned int N = 0;
	unsigned int np = 0;
	int StartFrom = 0;
	int EndBefore = 0;
	DataProcMapList thisStage, nextStage;
	DataProcMapList::iterator group;

	N = getNumberOfElements();
	np = getNumberOfProcessors();
	nStages = (unsigned int)ceil(log(N)/log(4));
	Limit = new unsigned int[np];
	for (i=0; i<np; i++){
		Limit[i] = np;
	}
	thisStage.push_back(CDataProcMap(0, N, 0));

	if (np>1){
		for (i=0; i<nStages-1; i++){
			// Reverse Four-way Tree (with the root node in the top)
			for (group=thisStage.begin(); group!=thisStage.end(); group++){
				DataProcMapList tempList, nextTempList;
				DataProcMapList::iterator tempListIterator;
				tempList.push_back(*group);
				for (unsigned int r=0; r<2; r++){
					for (tempListIterator=tempList.begin(); tempListIterator!=tempList.end(); tempListIterator++){
					j = (*tempListIterator).Processor;
						k = j + (unsigned int)ceil((Limit[j]-j)/2);
						if (k>=Limit[j]){
							k = Limit[j] - 1;
						}
						StartFrom = (*tempListIterator).StartFrom;
						EndBefore = (*tempListIterator).EndBefore;
						if (k!=j){
							// Compute the next level of Limits
							Limit[k] = Limit[j];
							Limit[j] = k;
							// Divide this group into two divisions
							nextTempList.push_back(CDataProcMap(StartFrom, (StartFrom+EndBefore)/2, j));
							nextTempList.push_back(CDataProcMap((StartFrom+EndBefore)/2, EndBefore, k));
						}
						else{
							// No division (no processors left)
							nextTempList.push_back(CDataProcMap(StartFrom, EndBefore, j));
						}
					}
					tempList.swap(nextTempList);
					nextTempList.clear();
				}	
				for (tempListIterator=tempList.begin(); tempListIterator!=tempList.end(); tempListIterator++){
					nextStage.push_back(*tempListIterator);
				}
				tempList.clear();
			}
			// In the next stage, the current nextStage will be thisStage
			thisStage.swap(nextStage);
			// Clear nextStage so that we have a clean list
			nextStage.clear();
		}
	}

	// Copy the result to the output
	lastStage.swap(thisStage);
	thisStage.clear();

#if(DEBUG==1)
	for (group=lastStage.begin(); group!=lastStage.end(); group++){
		cout << "[" << (*group).Processor << "]" << (*group).StartFrom << " .. " << (*group).EndBefore << endl;
	}
#endif

	return true;
}

bool Radix4FFT::UniProcessorFFT(){
	unsigned int nStages = 0;
	unsigned int nDivisions = 0;
	unsigned int nButterfliesPerStage = 0;
	unsigned int N = 0;
	unsigned int nElementsPerDivision = 0;
	unsigned int nButterfliesPerDivision = 0;
	unsigned int nElementsInNextDivision = 0;
	unsigned int StartFrom = 0, EndBefore = 0;
	int rank;
	int np;
	unsigned int i = 0, j = 0, k = 0;
	ElementType *Data = NULL;

	// This does not support decimation-in-time yet
	if (getDecimationInTime()==true){
		return false;
	}

	// Setup some initial values
	N = getNumberOfElements();
	nStages = (unsigned int)(ceil(log(N)/log(4)));
	nButterfliesPerStage = N/4;
	nDivisions = 1;
	rank = MPI::COMM_WORLD.Get_rank();
#if(DEBUG==1)
	if (rank==0){
		cout << "Number Of Stages: " << nStages << endl;
	}
#endif
	np = getNumberOfProcessors();
	Data = getInElements();

	// Perform the computation
	for (i=0; i<nStages; i++){
		nElementsPerDivision = N / nDivisions;
		for (unsigned int j=0; j< nDivisions; j++){
			StartFrom = j*nElementsPerDivision;
			EndBefore = StartFrom + nElementsPerDivision;
			FFTOnDivision(StartFrom, EndBefore);
		}
		nDivisions *= 4;
	}

	return true;
}

bool Radix4FFT::mpiFFT(){
	unsigned int nStages = 0;
	unsigned int nDivisions = 0;
	unsigned int nButterfliesPerStage = 0;
	unsigned int N = 0;
	unsigned int nElementsPerDivision = 0;
	unsigned int nButterfliesPerDivision = 0;
	unsigned int nElementsInNextDivision = 0;
	unsigned int StartFrom = 0, EndBefore = 0;
	unsigned int myStartFrom = 0, myEndBefore = 0;
	int rank;
	int np;
	unsigned int i = 0, j = 0, k = 0;
	ElementType *Data = NULL;
	DataProcMapList dataProcList;
	DataProcMapList::iterator group;
	DataStageMapList thisStage, nextStage;
	DataStageMapList::iterator thisGroup;

	// This does not support decimation-in-time yet
	if (getDecimationInTime()==true){
		return false;
	}

	// Setup some initial values
	N = getNumberOfElements();
	nStages = (unsigned int)(ceil(log(N)/log(4)));
	nButterfliesPerStage = N/4;
	rank = MPI::COMM_WORLD.Get_rank();
#if (DEBUG==1)
	if (rank==0){
		cout << "Number Of Stages: " << nStages << endl;
	}
#endif
	np = getNumberOfProcessors();
	Data = getInElements();
	thisStage.push_back(CDataStageMap(0, N));

	// Decide on who should communicate with whom
	LayOutDataProcMap(dataProcList);
	for (group = dataProcList.begin(); group!=dataProcList.end(); group++)
	{
		i = (*group).Processor;
		if (i==rank){
			myStartFrom = (*group).StartFrom;
			myEndBefore = (*group).EndBefore;
		}
	}

	// Communicate, only if we have more than one processor
	if (np>1){
		// Wait for everybody to join us
		MPI::COMM_WORLD.Barrier();
		// Now Broadcast the input data
		MPI::COMM_WORLD.Bcast((void*)(Data), N*sizeof(ElementType)/sizeof(char), MPI::CHAR, 0);
	}

	// Perform the computation
	for (i=0; i<nStages; i++){
		for (thisGroup=thisStage.begin(); thisGroup!=thisStage.end(); thisGroup++){
			StartFrom = (*thisGroup).StartFrom;
			EndBefore = (*thisGroup).EndBefore;
			if ((StartFrom<=myStartFrom && myStartFrom<EndBefore) ||
				(StartFrom<myEndBefore && myEndBefore<=EndBefore) ||
				(myStartFrom<=StartFrom && StartFrom<myEndBefore) ||
				(StartFrom<=myStartFrom && myEndBefore<=EndBefore)
				){
#if(DEBUG==1)
				cout << "[" << rank << "]" << i << ", " << StartFrom << ", " << EndBefore << " (" << myStartFrom << "," << myEndBefore << ")" << endl;
#endif
				// This processor should process atleast part of this group to proceed
				FFTOnDivision(StartFrom, EndBefore);
				// Chart out the next DataStageMapList based on this stage
				if (i!=(nStages-1)){
					nElementsInNextDivision = (EndBefore - StartFrom)/4;
					for (k=0; k<4; k++){
						nextStage.push_back(CDataStageMap(StartFrom+k*nElementsInNextDivision, StartFrom+(k+1)*nElementsInNextDivision));
					}
				}
			}
		}
		thisStage.swap(nextStage);
		nextStage.clear();
	}

	// Prepare for communication and then communicate
	if (np>1){
		// Get all the data back to proc 0 if this is decimation in frequency
		int *Count = NULL;
		int	*Displ = NULL;
		Count = new int [np];
		Displ = new int [np];
		for (i=0; i<np; i++){
			Count[i] = 0;
			Displ[i] = 0;
		}	
		for (group = dataProcList.begin(); group!=dataProcList.end(); group++)
		{
			i = (*group).Processor;
			StartFrom = (*group).StartFrom;
			EndBefore = (*group).EndBefore;
			Count[i] = (EndBefore-StartFrom)*sizeof(ElementType)/sizeof(char);
			Displ[i] = StartFrom*sizeof(ElementType)/sizeof(char);
		}

		// Communicate, only if we have more than one processor
		// Wait for everybody to join us
		MPI::COMM_WORLD.Barrier();
		// Processor 0 to gather the result and compose final data
		MPI::COMM_WORLD.Gatherv((const void*)(Data+myStartFrom), Count[rank], MPI::CHAR, (void*)(Data), Count, Displ, MPI::CHAR, 0);
		//Reshuffle(getResult());
	}
	dataProcList.clear();

	return true;
}

bool Radix4FFT::performFFT(){
	bool bDone = false;
	double timeTaken = 0;
	int rank = 0;

	rank = MPI::COMM_WORLD.Get_rank();

	if (rank==0){
		timeTaken = MPI::Wtime();
	}

	if (getNumberOfProcessors()==1){
		bDone = UniProcessorFFT();
	}
	else{
		bDone = mpiFFT();
	}

	if (rank==0){
		timeTaken = MPI::Wtime() - timeTaken;
		cout << "Time taken for " << getNumberOfElements() << " elements using " << getNumberOfProcessors() << " processors = " << timeTaken << " seconds" << endl;
	}

	return bDone;
}

/*
	The main function
	This initialises the various part of the code and leaves the rest to the classes.
*/

int main(int argc, char **argv){
  int rank, size;
  FFT *fft;
  int nElements = 16;
  string resultStr("The spectrum is");
  string inputStr("The inputs are");

  MPI_Init(&argc, &argv);
  size = MPI::COMM_WORLD.Get_size();
  rank = MPI::COMM_WORLD.Get_rank();

  if (argc>1){
	  nElements = atoi(argv[1]);
  }
  fft = new Radix4FFT(nElements, size);
  if (rank==0){
	for (unsigned int i=0; i<nElements; i++){
		fft->setInElements(i, ElementType(rand()/nElements, rand()/nElements));
	}
  }
  //fft->printInElements(inputStr);
  fft->setDecimationInTime(false);
  fft->performFFT();
  //fft->printResult(resultStr);
  delete fft; fft = NULL;

  MPI_Finalize();
}

